package cn.edu.zjvtit.cloudstorage.apiservice.controller;

import cn.edu.zjvtit.cloudstorage.apiservice.ApiServiceApplication;
import cn.edu.zjvtit.cloudstorage.apiservice.entity.ObjectEntity;
import cn.edu.zjvtit.cloudstorage.apiservice.rabbitmq.topic.Sender;
import cn.edu.zjvtit.cloudstorage.apiservice.repository.ObjectRepository;
import cn.edu.zjvtit.cloudstorage.apiservice.store.DataServerStore;
import com.backblaze.erasure.ByteInputOutputExpCodingLoop;
import com.backblaze.erasure.CodingLoop;
import com.backblaze.erasure.ReedSolomon;
import com.rabbitmq.client.Channel;
import org.slf4j.Logger;
import org.springframework.amqp.core.BindingBuilder;
import org.springframework.amqp.core.Message;
import org.springframework.amqp.core.Queue;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageRequest;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.util.DigestUtils;
import org.springframework.util.MultiValueMap;
import org.springframework.web.bind.annotation.*;
import cn.edu.zjvtit.cloudstorage.apiservice.result.Result;
import org.springframework.web.client.RestTemplate;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;

import static java.util.Arrays.copyOfRange;

@RestController
public class ObjectsController {
    @Autowired
    private DataServerStore dataServerStore;
    @Autowired
    private Sender sender;

    @Autowired
    private Logger logger;
    @Autowired
    private RestTemplate restTemplate;
    @Autowired
    private RabbitTemplate rabbitTemplate;
    @Autowired
    private ObjectRepository objectRepository;

    @PutMapping("/objects/{filename}")
    public String putObject(@RequestBody byte[] requestBody,
                            @RequestHeader Map<String, String> headers,
                            @PathVariable("filename") String filename) throws IOException {
        String server = this.dataServerStore.selectRandomServer();
        logger.info("put object to server: " + server);
        if (server == null) {
            return new Result("no server available", "10000").toString();
        }
        String md5 = headers.get("md5");
        ObjectEntity object = this.objectRepository.findFirstByNameOrderByVersionDesc(filename);
        if (object != null) {
            if (md5.equals(object.getHash())) {
                return new Result("object already exists", "10001").toString();
            }
            Integer version = object.getVersion() + 1;
            object = new ObjectEntity();
            object.setName(filename);
            object.setHash(md5);
            object.setVersion(version);
            object.setSize(requestBody.length);
        } else {
            object = new ObjectEntity();
            object.setName(filename);
            object.setHash(md5);
            object.setVersion(1);
            object.setSize(requestBody.length);
        }

        int blockSize = (int) Math.ceil(requestBody.length / 4.0);
        ReedSolomon codec = ReedSolomon.create(4, 2);
        byte[][] shards = new byte[6][];
        shards[0] = copyOfRange(requestBody, 0, blockSize - 1);
        shards[1] = copyOfRange(requestBody, blockSize, 2 * blockSize - 1);
        shards[2] = copyOfRange(requestBody, 2 * blockSize, 3 * blockSize - 1);
        shards[3] = new byte[blockSize];
        for (int i = 3 * blockSize; i < requestBody.length - 1; i++) {
            shards[3][i - 3 * blockSize] = requestBody[i];
        }
        shards[4] = new byte[blockSize];
        shards[5] = new byte[blockSize];
        codec.encodeParity(shards, 0,blockSize);
        // todo storage shards

        this.objectRepository.save(object);
        String apiUrl = "http://" + server;
        MultiValueMap<String, String> haderWithMd5 = new HttpHeaders();
        ;
        haderWithMd5.add("md5", md5);
        // 获取uuid
        ResponseEntity<Result> ret = this.restTemplate.exchange(apiUrl + "/temp", HttpMethod.POST, new HttpEntity<>(null, haderWithMd5), Result.class);
        String uuid = (String) Objects.requireNonNull(ret.getBody()).getData().get("uuid");
        // 发送patch请求上传文件
        HttpEntity<byte[]> request = new HttpEntity<>(requestBody);
        ret = this.restTemplate.exchange(apiUrl + "/temp/" + uuid, HttpMethod.PATCH, request, Result.class);
        String retMd5 = (String) Objects.requireNonNull(ret.getBody()).getData().get("md5");
        if (Objects.equals(md5, retMd5)) {
            // 将临时文件转为正式文件
            this.restTemplate.exchange(apiUrl + "/temp/" + uuid, HttpMethod.PUT, new HttpEntity<>(null, haderWithMd5), Result.class);
            return new Result("success", "10000").toString();
        } else {
            // 删除文件
            this.restTemplate.exchange(apiUrl + "/temp/" + uuid, HttpMethod.DELETE, new HttpEntity<>(null, haderWithMd5), Result.class);
            return new Result("md5 not match", "10002").toString();
        }

    }


    @GetMapping("/objects/{filename}")
    public String getObject(
            HttpServletResponse response,
            @PathVariable("filename") String filename,
            @RequestParam("version") int version
    ) throws IOException {
        // 查询哪台服务器有这个文件
        // 读取文件
        ObjectEntity obj;
        if (version == 0) {
            obj = this.objectRepository.findFirstByNameOrderByVersionDesc(filename);
        } else {
            obj = this.objectRepository.findByVersionAndName(version, filename);
        }
        String queueName = obj.getHash();
        Channel channel = rabbitTemplate.getConnectionFactory().createConnection().createChannel(false);
        channel.queueDeclare(queueName, false, false, false, null);
        channel.queueBind(queueName, "exchange", "");
        sender.send(queueName);
        try {
            Object msg = rabbitTemplate.receiveAndConvert(queueName, 200);
            assert msg != null;
            logger.info(msg.toString());
            String url = "http://" + msg + "/objects/" + obj.getHash();

            ResponseEntity<String> res = this.restTemplate.getForEntity(url, String.class);
            response.reset();
            response.setContentType("application/octet-stream");
            response.setCharacterEncoding("utf-8");
            response.setContentLength((int) res.getHeaders().getContentLength());
            response.setHeader("Content-Disposition", "attachment;filename=" + filename);
            OutputStream os = response.getOutputStream();
            os.write(res.getBody().getBytes());
            os.close();
            return "";
        } catch (Exception e) {
            logger.info("locate object error", e);
            return new Result("file not exists", "10000").toString();
        }
    }

    @GetMapping("/locate/{filename}")
    public String locateObject(@PathVariable("filename") String filename) throws IOException {
        // 查询哪台服务器有这个文件
        String queueName = filename;
        Channel channel = rabbitTemplate.getConnectionFactory().createConnection().createChannel(false);
        channel.queueDeclare(queueName, false, false, false, null);
        channel.queueBind(queueName, "exchange", "");
        sender.send(queueName);
        try {
            Object msg = rabbitTemplate.receiveAndConvert(queueName, 200);
            assert msg != null;
            logger.info(msg.toString());
            return new Result("file in server:" + msg.toString(), "10000").toString();
        } catch (Exception e) {
            logger.info("locate object error", e);
            return new Result("file not exists", "10000").toString();
        }

    }

    @GetMapping("/versions/{filename}")
    public String getVersions(@PathVariable("filename") String filename, @RequestParam("offset") int offset, @RequestParam("limit") int limit) {
        if (offset <= 0) {
            offset = 0;
        }
        if (limit <= 0) {
            limit = 10;
        }
        Page<ObjectEntity> result = this.objectRepository.findAllByNameOrderByVersionDesc(filename, PageRequest.of(offset, limit));
        Map<String, Object> map = new HashMap<>();
        map.put("total", result.getTotalElements());
        map.put("rows", result.getContent());
        return new Result("success", "10000").withData(map).toString();
    }

    @GetMapping("/versions")
    public String getVersionList(@RequestParam("offset") int offset, @RequestParam("limit") int limit) {
        if (offset <= 0) {
            offset = 0;
        }
        if (limit <= 0) {
            limit = 10;
        }
        Page<ObjectEntity> result = this.objectRepository.findAll(PageRequest.of(offset, limit));
        Map<String, Object> map = new HashMap<>();
        map.put("total", result.getTotalElements());
        map.put("rows", result.getContent());
        return new Result("success", "10000").withData(map).toString();
    }
}
